// Copyright (c) 2017  Anna Bigatti, Alberto Griggio
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"

///// first prototype
#include "CoCoA/ExternalLibs-MathSAT.H"
#include <gmpxx.h>

using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "This program shows the first prototype for MathSAT communication.\n";


//----------------------------------------------------------------------

namespace CoCoA
{

typedef std::pair<BigRat, int> Monomial; // <coefficient, variable index>

class LinearConstraint {
public:
    enum RelOp {
        EQUAL,
        NOTEQUAL,
        LEQ,
        LT
    };
  explicit LinearConstraint(RelOp r=EQUAL): rel_(r) {}
    void add_var(const BigRat &coeff, int var)
    { lhs_.push_back(std::make_pair(coeff, var)); }
    void add_const(const BigRat &coeff) { rhs_ += coeff; }

    typedef std::vector<Monomial>::const_iterator const_iterator;
    const_iterator begin() const { return lhs_.begin(); }
    const_iterator end() const { return lhs_.end(); }
    const BigRat &coeff() const { return rhs_; }
    RelOp rel() const { return rel_; }
private:
    RelOp rel_;
    std::vector<Monomial> lhs_;
    BigRat rhs_;
};


  std::string to_string(const BigRat &r)
  {
    std::ostringstream buf;
    buf << r;
    return buf.str();
  }
  

  msat_term var_to_mathsat(msat_env env, int var)
  {
    std::ostringstream buf;
    buf << "X" << var;
    std::string name = buf.str();
    msat_type rat = msat_get_rational_type(env);
    msat_decl d = msat_declare_function(env, name.c_str(), rat);
    msat_term x = msat_make_constant(env, d);
    return x;
  }

  int get_var_index(msat_term x)
  {
    msat_decl d = msat_term_get_decl(x);
    char *name = msat_decl_get_name(d);
    int var = atoi(name+1);
    msat_free(name);
    return var;
  }


msat_term to_msat(msat_env env, const LinearConstraint &lc)
{
    msat_term lhs = msat_make_number(env, "0");
    std::string s = to_string(lc.coeff());
    msat_term rhs = msat_make_number(env, s.c_str());

    for (LinearConstraint::const_iterator it = lc.begin(), end = lc.end();
         it != end; ++it) {
        s = to_string(it->first);
        msat_term c = msat_make_number(env, s.c_str());
        int v = it->second;
        msat_term x = var_to_mathsat(env, v);
        lhs = msat_make_plus(env, lhs, msat_make_times(env, c, x));
    }

    switch (lc.rel()) {
    case LinearConstraint::EQUAL:
        return msat_make_equal(env, lhs, rhs);
    case LinearConstraint::NOTEQUAL:
        return msat_make_not(env, msat_make_equal(env, lhs, rhs));
    case LinearConstraint::LEQ:
        return msat_make_leq(env, lhs, rhs);
    default: // case LinearConstraint::LT
        return msat_make_not(env, msat_make_leq(env, rhs, lhs));
    }
}


  void program()
  {
    GlobalManager CoCoAFoundations;

#ifndef CoCoA_WITH_MATHSAT
    cout << "MathSAT library is not available to CoCoALib." << endl;
#else // MathSAT is available
    cout << ShortDescription << endl;
    cout << boolalpha; // so that bools print out as true/false


    LinearConstraint c1(LinearConstraint::LEQ), c2(LinearConstraint::LEQ), c3(LinearConstraint::LEQ), c4(LinearConstraint::LEQ);
    // X0 <= 3/10
    c1.add_var(BigRat(1,1), 0);
    c1.add_const(BigRat(3, 10));
    // -X0 <= 0
    c2.add_var(BigRat(-1,1), 0);
    // X1 + 3*X0 <= 7/10
    c3.add_var(BigRat(1,1), 1); // 1 * X1
    c3.add_var(BigRat(3,1), 0); // 3 * X0
    c3.add_const(BigRat(7, 10));
    // -X1 <= -1/2
    c4.add_var(BigRat(-1,1), 1);
    c4.add_const(BigRat(-1, 2));

    // create a MathSAT environment
    msat_config cfg = msat_create_config();
    msat_set_option(cfg, "model_generation", "true");
    msat_env env = msat_create_env(cfg);
    msat_destroy_config(cfg);

    // add the constraints
    msat_assert_formula(env, to_msat(env, c1));
    msat_assert_formula(env, to_msat(env, c2));
    msat_assert_formula(env, to_msat(env, c3));
    msat_assert_formula(env, to_msat(env, c4));

    // check feasibility
    msat_result res = msat_solve(env);

    switch (res) {
    case MSAT_SAT: {
        std::cout << "FEASIBLE, SOLUTION IS:" << std::endl;
        matrix M = NewDenseMat(RingQQ(), 1, 2); // 2 = numero di variabili
        BigRat ans;
        msat_model_iterator it = msat_create_model_iterator(env);
        while (msat_model_iterator_has_next(it)) {
            msat_term t, v;
            msat_model_iterator_next(it, &t, &v);
            int var = get_var_index(t);
            msat_term_to_number(env, v, mpqref(ans));
            SetEntry(M, 0, var, ans);
        }
        cout << M << endl;
        
    }   break;
    case MSAT_UNSAT:
        std::cout << "INFEASIBLE" << std::endl;
        break;
    default:
        std::cout << "something bad happened! " << msat_last_error_message(env)
                  << std::endl;
    }

    msat_destroy_env(env);
#endif // CoCoA_WITH_MATHSAT
  }




} // end of namespace CoCoA

//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    CoCoA::program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    cerr << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(cerr, err);
  }
  catch (const std::exception& exc)
  {
    cerr << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    cerr << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }

  CoCoA::BuildInfo::PrintAll(cerr);
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-MathSat1.C,v 1.1 2017/02/24 08:19:39 bigatti Exp $
// $Log: ex-MathSat1.C,v $
// Revision 1.1  2017/02/24 08:19:39  bigatti
// -- first import
//
// Revision 1.26  2015/09/03 10:17:11  bigatti
// -- changes by Christof Soeger (in Aarhus)
//
// Revision 1.25  2015/06/29 15:27:25  bigatti
// -- minor fix
//
// Revision 1.24  2015/06/29 12:20:43  bigatti
// -- moved in cocoa namespace
//
// Revision 1.23  2014/07/11 10:11:53  abbott
// Summary: Christof's cleaning of the example
// Author: JAA
//
// Revision 1.22  2014/01/30 09:55:55  bigatti
// -- removed DOS newlines
//
// Revision 1.21  2014/01/29 17:34:22  bigatti
// -- added HilbertSeries (by Christof Soeger)
//
// Revision 1.20  2012/10/08 13:53:45  bigatti
// -- more cleaning and updates by Christof Soeger
//
// Revision 1.19  2012/08/03 16:31:22  bigatti
// -- changed: procedural --> functional (by C.Soeger)
//
// Revision 1.18  2012/07/20 13:27:33  abbott
// Commented out the code which uses the (obsolescent) pseudo-ctors NewConeLong
// and NewConeBigInt.
//
// Revision 1.17  2012/07/19 17:11:13  abbott
// Added unified "NewCone" (user does not have to choose between long or BigInt).
// Cleaned up.
//
// Revision 1.16  2012/07/04 12:31:08  abbott
// Changed int into long; improved ShortDescription & LongDescription.
//
// Revision 1.15  2012/06/19 14:39:39  bigatti
// -- unix-ified
//
// Revision 1.14  2012/06/19 14:36:23  bigatti
// -- by C.Soeger
//
// Revision 1.13  2011/11/09 15:13:49  bigatti
// -- new syntax for HilbertBasis
//
// Revision 1.12  2011/10/12 15:50:17  abbott
// Simplified use of CPP macro CoCoA_WITH_NORMALIZ.
//
// Revision 1.11  2011/09/30 12:55:44  bigatti
// -- introduced namespace "Normaliz" and removed Normaliz from function names
// -- input of Normaliz functions in CoCoA-5 is now a matrix instead of
//    vector<vector<BigInt>>
//
// Revision 1.10  2011/08/23 12:04:04  bigatti
// -- updated after renaming ZZ --> BigInt
//
// Revision 1.9  2011/07/20 13:49:37  bigatti
// -- added "Normaliz" postfix to Normaliz function calls
//
// Revision 1.8  2011/07/20 12:45:12  bigatti
// -- new normaliz interface (not yet public)
//
// Revision 1.7  2011/02/17 16:50:04  bigatti
// -- getting ready for new official veson on Normaliz: added HVector, removed Triangulation
//
// Revision 1.6  2010/12/17 16:07:54  abbott
// Ensured that all i/o in examples is on standard C++ streams
// (rather than GlobalInput(), etc).
//
// Revision 1.5  2010/11/22 17:33:31  bigatti
// -- input with c-style array
//
// Revision 1.4  2010/10/12 11:22:54  bigatti
// -- TmpNormaliz.H simplified:
//    now Cone is a smart pointer to ConeBase
//    and the concrete classes are entirely in the .C file
// -- added Ht1Elements, SupportHyperplanes, Triangulation
// -- added some text in ex-Normaliz1 and 2
//
// Revision 1.3  2010/10/08 13:53:51  bigatti
// -- cleaned up
//
// Revision 1.2  2010/10/08 10:39:32  bigatti
// -- extended interface for normaliz
//
// Revision 1.1  2010/10/05 14:31:26  bigatti
// -- first import
//
// Revision 1.4  2008/10/07 12:12:54  abbott
// Removed useless commented out #include.
//
// Revision 1.3  2007/05/31 16:06:16  bigatti
// -- removed previous unwanted checked-in version
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.9  2007/03/07 11:51:40  bigatti
// -- improved test alignment
//
// Revision 1.8  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.7  2007/03/02 17:46:40  bigatti
// -- unique RingZ and RingQ
// -- requires foundations.H ;  foundations blah;  (thik of a better name)
//
// Revision 1.6  2007/03/02 10:47:53  cocoa
// First stage of RingZ modifications -- tests do not compile currently, Anna will fix this.
//
// Revision 1.5  2007/03/01 13:52:59  bigatti
// -- minor: fixed typo
//
// Revision 1.4  2007/02/28 15:15:56  bigatti
// -- minor: removed quotes in description
//
// Revision 1.3  2007/02/12 16:27:43  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.2  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.1  2006/03/12 21:28:34  cocoa
// Major check in after many changes
//
